package org.ids_mannheim;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.concurrent.*;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import com.vdurmont.etaprinter.ETAPrinter;
import picocli.CommandLine;

@CommandLine.Command(mixinStandardHelpOptions = true,
        name = "totalngram", description = "add ngram counts from KorAP-XML or CoNLL-U files")

public class TotalNGram implements Callable<Integer> {
    private static final int MAX_THREADS = Runtime.getRuntime().availableProcessors() * 2 / 3;
    private ETAPrinter etaPrinter;

    @CommandLine.Option(names = {"-L",
            "--log-file"}, defaultValue = "sum.log", description = "log file name (default: ${DEFAULT-VALUE})")
    String logFileName;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"-o",
            "--output-file"}, description = "Output file (default: -")
    File output_file = null;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"-P",
            "--max-procs"}, description = "Run up to max-procs processes at a time (default: ${DEFAULT-VALUE})")
    int max_threads = MAX_THREADS;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"-n",
            "--ngram-size"}, description = "n-gram size (default: ${DEFAULT-VALUE})")
    int ngram_size = 1;

    @SuppressWarnings("CanBeFinal")
    @CommandLine.Option(names = {"-f",
            "--folds"}, description = "number of folds (default: ${DEFAULT-VALUE})")
    int FOLDS = 10;

    @CommandLine.Parameters(arity = "1..*", description = "input files")
    private final ArrayList<String> inputFiles = new ArrayList<>();

    public TotalNGram() {
    }

    @Override
    public Integer call() throws Exception {
        PrintStream output_stream = (output_file == null || output_file.equals("-") ? System.out : new PrintStream(output_file));

        FoldedEntry.setFolds(FOLDS);
        ConcurrentHashMap<String, FoldedEntry> map = new ConcurrentHashMap<>();
        long totalFilesSizes = inputFiles.parallelStream().mapToLong(fname -> new File(fname).length()).sum();
        etaPrinter = ETAPrinter.init("sum", totalFilesSizes, System.err, false);
        BlockingQueue<Integer> queue = new LinkedBlockingQueue<>(inputFiles.size());
        ExecutorService es = Executors.newCachedThreadPool();
        int threads = Math.min(max_threads, inputFiles.size());
        IntStream.range(0, threads).forEach(unused -> es.execute(new Worker(queue, inputFiles, ngram_size, FOLDS, map, etaPrinter)));
        queue.addAll(IntStream.range(0, inputFiles.size()).boxed().collect(Collectors.toList()));
        IntStream.range(0, threads).forEach(unused -> {
            try {
                queue.put(-1);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        });
        es.shutdown();
        boolean finished = es.awaitTermination(120, TimeUnit.HOURS);
        map.entrySet().stream()
                .sorted(Collections.reverseOrder(new ValueThenKeyComparator<>()))
                .forEach(entry -> output_stream.println(entry.getKey() + entry.getValue().toString()));
        IntStream.rangeClosed(1, FOLDS)
                .forEach(i -> output_stream.print("\t" + map.values()
                        .parallelStream().mapToLong(e -> e.count.get(i)).sum()));
        output_stream.println("\t" + map.values().parallelStream().mapToLong(e -> e.count.get(0)).sum());
        return null;
    }

    public static void main(String[] args) throws FileNotFoundException {
        System.exit(new CommandLine(new TotalNGram()).execute(args));
    }
}
